"""
The following functions implement the Reverse Local Push on 
Dynamic graphs. 
Resource: "Approximate Personalized PageRank on Dynamic Graphs"
https://www.kdd.org/kdd2016/papers/files/rfp1146-zhangA.pdf
"""

def update_reverse_push(graph, u, v, p, r, double alpha, action='d'):
    """
    Updates the r[u] caused by deletion or addition of (u, v)
    before the actual deletion/addition 
    """
    tmp = ((1 - alpha) * p.get(v, 0) - p.get(u, 0) - alpha * r.get(u, 0)) * graph[u][v]['weight']
    if u == v:
        tmp += 1

    if action == 'd':
        new_denom = 1 - graph[u][v]['weight']
        r[u] -= float(tmp) / (alpha * new_denom)
    else:
        new_denom = 1 + graph[u][v]['weight']
        r[u] += float(tmp) / (alpha * new_denom)
    return r

cdef double compute_delta_r(graph, u, neighbors, p, r, double alpha):
    """
    Computes the delta r when multiple edges are deleted.
    The implementation is based on the update rule in Algorithm 3
    in the paper.
    """
    cdef double sum_weights, avg_deleted, old_avg, tmp

    neighbors_weights = [graph[u][v]['weight'] for v in neighbors]
    sum_weights = sum(neighbors_weights)
    avg_deleted = sum([p.get(neighbors[i], 0) * neighbors_weights[i] for i in range(len(neighbors))])
    avg_deleted /= sum_weights

    old_avg = (p.get(u, 0) + alpha * r.get(u, 0)) / (1 - alpha)
    tmp = (avg_deleted - old_avg) * sum_weights / (1.0 - sum_weights)

    return (float(tmp) * (1 - alpha)) / alpha

cdef double compute_delta_r_forward(graph, u, p, double alpha):
    cdef double edge_weight
    edge_weight = graph[u][0]['weight']
    return p * edge_weight * (1 / alpha)

cdef double compute_delta_rn_forward(graph, u, neighbors, p, r, double alpha):
    cdef double edge_weight
    edge_weight = graph[u][0]['weight']
    return ((1 - alpha) * p * edge_weight) * (1 / alpha)

cdef double compute_delta_p(graph, u):
    cdef double edge_weight
    edge_weight = graph[u][0]['weight']

    return (edge_weight - 1 / edge_weight)

def update_reverse_push_multi(graph, u, neighbors, p, r, double alpha, action='d'):
    cdef double delta_r

    delta_r = compute_delta_r(graph, u, neighbors, p, r, alpha)
    if action == 'd':
        r[u] = r.get(u, 0) - delta_r
    else:
        r[u] = r.get(u, 0) + delta_r
    return r

def update_forward_push_multi(graph, source_node, neighbors, p, r, double alpha, double e, action='d'):
    if len(neighbors) > 0:
        if action == 'd':
            edge_deletion_multiple(graph, source_node, neighbors)
            deg = graph.out_degree(source_node)
            delta_p = deg / (deg + 1)
            p[source_node] = p.get(source_node, 0) * delta_p
            try:
                delta_r = (p.get(source_node, 0) / deg) * (1 / alpha)
            except ZeroDivisionError:
                delta_r = 0
            r[source_node] = r.get(source_node, 0) + delta_r
            delta_rn = {}
            for v in neighbors:
                try:
                    delta_rn[v] = ((((1 - alpha) * p.get(source_node, 0)) / deg) * (1 / alpha))
                except ZeroDivisionError:
                    delta_rn[v] = 0
                r[v] = r.get(v, 0) - delta_rn[v]
        else:
            edge_addition_multiple(graph, source_node, neighbors)
            deg = graph.out_degree(source_node)
            if deg > 1:
                delta_p = deg / (deg - 1)
                p[source_node] = p.get(source_node, 0) * delta_p
                delta_r = (p.get(source_node, 0) / deg) * (1 / alpha)
                r[source_node] = r.get(source_node, 0) - delta_r
                delta_rn = {}
                for v in neighbors:
                    delta_rn[v] = (((1 - alpha) * p.get(source_node, 0) / deg) * (1 / alpha))
                    r[v] = r.get(v, 0) + delta_rn[v]

    return forward_local_push(graph, source_node, p, r, alpha, e, update=True)

def forward_local_push(graph, s, p, r, double alpha=0.15, double e=0.00008, bint update=False):
    """
    This function implements Algorithm 1 in the paper
    """
    cdef int deg
    cdef double edge_weight
    in_r = []
    if not update:
        p = {}
        r = {s: 1.0}
    for k, v in r.items():
        if abs(float(v)) > e:
            in_r.append(k)
    while in_r:
        push_node = in_r.pop()
        deg = graph.out_degree(push_node)
        deg = deg if deg != 0 else 1
        if abs(float(r[push_node])) / (deg) <= e:
            continue

        p[push_node] = p.get(push_node, 0) + alpha * r[push_node]
        for node in graph.successors(push_node):
            if node == push_node:
                continue

            edge_weight = graph[push_node][node]['weight']

            r[node] = r.get(node, 0) + (1 - alpha) * r[push_node] * edge_weight
            if abs(float(r.get(node, 0)) * edge_weight) > e:
                in_r.append(node)

        r[push_node] = 0
    return p, r

def reverse_local_push(graph, t, p, r, double alpha=0.15, double e=0.00008, bint update=False):
    """
    This function implements Algorithm 2 in the paper
    """

    out_r = []
    if not update:
        p = {}
        r = {t: 1.0}
    for k, v in r.items():
        if abs(float(v)) > e:
            out_r.append(k)
    iter = 0
    while out_r:
        iter += 1
        push_node = out_r.pop()
        if abs(float(r[push_node])) <= e:
            continue
        p[push_node] = p.get(push_node, 0) + alpha * r[push_node]
        self_weight = 0
        if graph.has_edge(push_node, push_node):
            self_weight = graph[push_node][push_node]['weight']
        for node in graph.predecessors(push_node):
            if node == push_node:
                continue
            edge_weight = graph[node][push_node]['weight']
            r[node] = r.get(node, 0) + (1 - alpha) * r[push_node] * edge_weight
            if abs(float(r.get(node, 0))) > e:
                out_r.append(node)

        r[push_node] = (1 - alpha) * self_weight * r[push_node]
        if abs(float(r[push_node])) > e:
            out_r.append(push_node)
    return p, r

def edge_deletion(graph, u, v):
    delta_weight = graph[u][v]['weight']
    for node in graph.successors(u):
        graph[u][node]['weight'] /= (1 - delta_weight)
    graph.remove_edge(u, v)

def edge_deletion_multiple(graph, u, neighbors):
    weights = [graph[u][v]['weight'] for v in neighbors]
    sum_weights = sum(weights)
    for n in neighbors:
        graph.remove_edge(u, n)
    for n in graph.successors(u):
        graph[u][n]['weight'] /= (1.0 - sum_weights)
    return weights

def edge_addition(graph, u, v, weight):
    for node in graph.successors(u):
        graph[u][node]['weight'] *= (1 - weight)
    graph.add_edge(u, v)
    graph[u][v]['weight'] = weight

def edge_addition_multiple(graph, u, neighbors, weights=None):
    if weights == None:
        graph.add_edges_from([(u, n) for n in neighbors])
        deg = graph.out_degree(u)
        for n in graph.successors(u):
            graph[u][n]['weight'] = (1.0 / deg)
    else:
        sum_weights = sum(weights)
        for n in graph.successors(u):
            graph[u][n]['weight'] *= (1.0 - sum_weights)
        index = 0
        for n in neighbors:
            graph.add_edge(u, n)
            graph[u][n]['weight'] = weights[index]
            index += 1

def update_graph_mix(g, source_node, removed_neighbors, added_neighbors, alpha, epsilon, pf, rf):
    if pf and rf:
        p, r = update_forward_push_multi(g, source_node, removed_neighbors, pf, rf, alpha, epsilon,
                                         action="d")
    else:
        edge_deletion_multiple(g, source_node, removed_neighbors)
        p, r = forward_local_push(g, source_node, {}, {}, alpha=alpha, e=epsilon)

    if pf and rf:
        p, r = update_forward_push_multi(g, source_node, added_neighbors, pf, rf, alpha, epsilon, action="a")
    else:
        edge_addition_multiple(g, source_node, added_neighbors)
        p, r = forward_local_push(g, source_node, {}, {}, alpha=alpha, e=epsilon)
    return g, p, r